// Copyright (C) 2019  Wilko Manger
//
// This file is part of matrix-dart-sdk-sqflite.
//
// matrix-dart-sdk-sqflite is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk-sqflite is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk-sqflite.  If not, see <https://www.gnu.org/licenses/>.

library matrix_sdk_sqflite;

import 'dart:convert';

import 'package:matrix_sdk/matrix_sdk.dart';
import 'package:meta/meta.dart';
import 'package:sqflite/sqflite.dart';

class SqfliteStore implements Store {

  final String path;
  SqfliteStore({@required this.path});

  bool _isOpen = false;

  @override
  bool get isOpen => _isOpen;

  LocalUser _user;

  Database _db;

  static const _userInfoTable = 'user_info';
  static const _roomTable = 'room';
  static const _eventTable = 'event';

  static const _idColumn = 'id';

  static const _userIdColumn = 'user_id';
  static const _accessTokenColumn = 'access_token';
  static const _syncTokenColumn = 'sync_token';
  static const _homeserverColumn = 'homeserver';

  // Room columns
  static const _nameColumn = 'name';
  static const _typeColumn = 'type';

  // Event columns
  static const _roomIdColumn = 'room_id';
  static const _contentColumn = 'content';

  @override
  Future<bool> open({LocalUser user}) async {

    try {
      _db = await openDatabase(
        path,
        version: 1,
        onCreate: (Database db, int version) async {

          if (user == null) {
            throw ArgumentError(
                "user cannot be null if the database does not exist"
            );
          }

          _user = user;

          await db.execute(
          '''
          CREATE TABLE $_userInfoTable (
          $_idColumn INTEGER PRIMARY KEY,
          $_userIdColumn TEXT,
          $_accessTokenColumn TEXT,
          $_syncTokenColumn TEXT,
          $_homeserverColumn TEXT);
          ''');

          await db.execute(
          '''
            CREATE TABLE $_roomTable (
              $_idColumn TEXT PRIMARY KEY,
              $_typeColumn TEXT,
              $_nameColumn TEXT NULL);
          ''');

          await db.execute(
          '''
            CREATE TABLE $_eventTable (
              $_idColumn TEXT PRIMARY KEY,
              $_roomIdColumn TEXT,
              $_contentColumn TEXT NOT NULL,
              FOREIGN KEY($_roomIdColumn) REFERENCES $_roomTable($_idColumn));
          ''');
        }
      );
    } on ArgumentError {
      return false;
    }

    _isOpen = true;
    return true;
  }

  @override
  Future<void> close() async {
    _isOpen = false;
  }

  Future<String> _getUserInfo(String column) async {
    var results = await _db.rawQuery(
      '''
      SELECT $column FROM $_userInfoTable WHERE $_idColumn = 1;
      '''
    );

    if (results.isNotEmpty) {
      return results.first[column];
    } else {
      return null;
    }
  }

  Future<void> _setUserInfo(String column, String value) async {
    var batch = _db.batch();
    batch.execute(
      '''
      UPDATE $_userInfoTable SET $column = ?;
      ''',
      [value]
    );

    batch.execute(
      '''
      INSERT OR IGNORE INTO $_userInfoTable($_idColumn, $column)
        values (?, ?);
      ''',
      [1, value]
    );

    await batch.commit(noResult: true);
  }

  @override
  Future<UserId> getUserId() async {
    return UserId(await _getUserInfo(_userIdColumn));
  }

  @override
  Future<void> setUserId(UserId userId) async {
    print('userId = ${userId.toString()}');
    await _setUserInfo(_userIdColumn, userId.toString());
  }

  @override
  Future<String> getAccessToken() async {
    return _getUserInfo(_accessTokenColumn);
  }

  @override
  Future<void> setAccessToken(String accessToken) async {
    await _setUserInfo(_accessTokenColumn, accessToken);
  }

  @override
  Future<String> getSyncToken() async {
    return _getUserInfo(_syncTokenColumn);
  }

  @override
  Future<void> setSyncToken(String syncToken) async {
    await _setUserInfo(_syncTokenColumn, syncToken);
  }

  @override
  Future<Homeserver> getHomeserver() async {
    return Homeserver(Uri.parse(await _getUserInfo(_homeserverColumn)));
  }

  @override
  Future<void> setHomeserver(Homeserver homeserver) async {
    await _setUserInfo(_homeserverColumn, homeserver.uri.toString());
  }

  @override
  Future<JoinedRoom> getJoinedRoom(RoomId id) async {
    var results = await _db.rawQuery(
      '''
      SELECT $_nameColumn FROM $_roomTable WHERE $_idColumn = ?;
      ''',
      [id.toString()]
    );

    if (results.isNotEmpty) {
      var first = results.first;
      var name = first[_nameColumn];

      return JoinedRoom(id, store: this, name: name);
    } else {
      return null;
    }
  }

  @override
  Future<void> setJoinedRoom(RoomId id, JoinedRoom room) async {
    var batch = _db.batch();
    batch.execute(
      '''
      UPDATE $_roomTable SET $_typeColumn = ?,
                             $_nameColumn = ?
                         WHERE $_idColumn = ?;
      ''',
      ['join', room.name, id.toString()]
    );

    batch.execute(
      '''
      INSERT OR IGNORE INTO $_roomTable($_idColumn,
                                        $_typeColumn,
                                        $_nameColumn)
        values (?, ?, ?);
      ''',
      [id.toString(), 'join', room.name]
    );

    await batch.commit(noResult: true);
  }

  @override
  Stream<JoinedRoom> getJoinedRooms() async* {
    var results = await _db.rawQuery(
      '''
      SELECT $_idColumn, $_nameColumn FROM $_roomTable;
      '''
    );

    if (results.isNotEmpty) {
      for (var result in results) {
        var id = RoomId(result[_idColumn]);
        var name = result[_nameColumn];

        yield JoinedRoom(id, store: this, name: name);
      }
    }
  }

  @override
  Future<Event> getEvent(RoomId roomId, EventId id) async {
    var results = await _db.rawQuery(
      '''
      SELECT $_contentColumn FROM $_eventTable WHERE $_idColumn = ?
                                               AND $_roomIdColumn = ?;
      ''',
      [id.toString(), roomId.toString()]
    );

    if (results.isNotEmpty) {
      var content = results.first[_contentColumn];
      return Event.fromJson(roomId, json.decode(content));
    } else {
      return null;
    }
  }

  @override
  Stream<Event> getEvents(RoomId roomId) async* {
    var results = await _db.rawQuery(
      '''
      SELECT $_contentColumn FROM $_eventTable WHERE $_roomIdColumn = ?;
      ''',
      [roomId.toString()]
    );

    if (results.isNotEmpty) {
      for (var result in results) {
        var content = result[_contentColumn];

        yield Event.fromJson(roomId, json.decode(content));
      }
    }
  }

  @override
  Future<void> addEvent(RoomId roomId, Event event) async {
    await _db.execute(
      '''
      INSERT OR IGNORE INTO $_eventTable($_idColumn,
                                         $_roomIdColumn, 
                                         $_contentColumn)
        values (?, ?, ?);
      ''',
      [event.id.toString(), roomId.toString(), json.encode(event.toJson())]
    );
  }

  @override
  Future<void> addEvents(RoomId roomId, List<Event> events) async {

    var query =
    '''
      INSERT OR IGNORE INTO $_eventTable($_idColumn,
                                         $_roomIdColumn,
                                         $_contentColumn)
     ''';

    var values = [];

    for (var event in events) {
      query += '\nvalues (?, ?, ?)';

      values.addAll([
        event.id.toString(), roomId.toString(), json.encode(event.toJson())]);
    }

    await _db.execute(query, values);
  }
}
